//   RW Jsonutils
// ================
// ______
// -The RW Foundational Libraries

// Filename: ./src/rw/json/json_util.cpp
// -Definitions of some baseline RW Jsonutils features

#include "incls_all.hpp"

namespace rw::json {

//------------------------------------------------------------------------------
auto fstream_2_json(std::fstream& json_ifs, bool const close_strm) -> Json {
  Json res{};

  try {
    json_ifs >> res;  // very clean tbh
  }
  // https://json.nlohmann.me/api/basic_json/operator_gtgt/#exceptions
  catch (Json::parse_error const& e) {
    // cerr << "\n  Json::parse_error \n'" << e.what()
    //      << "'\n  caught in fstream_2_json()\n\n";

    // // DESIGN: report stream position here as well?

    // if (e.id == 101)
    //   cerr << "'an unexpected token'\n"
    //        << "https://json.nlohmann.me/home/exceptions/"
    //           "#jsonexceptionparse_error101\n";
    // else if (e.id == 102)
    //   cerr << "'to_unicode fails or surrogate error.'\n"
    //        << "https://json.nlohmann.me/home/exceptions/"
    //           "#jsonexceptionparse_error102\n";
    // else if (e.id == 103)
    //   cerr << "'to_unicode fails'\n"
    //        << "https://json.nlohmann.me/home/exceptions/"
    //           "#jsonexceptionparse_error103\n";

    res.clear();

  } catch (exception const& e) {
    // cerr << "\n  exception '" << e.what() << "' caught in
    // fstream_2_json()\n\n";

    res.clear();
  }

  if (close_strm)
    json_ifs.close();

  return res;
}

//------------------------------------------------------------------------------
auto rd_json_file(std::string const& filepath) -> Json {
  fstream ifs{filepath, ios::in};
  return fstream_2_json(ifs);
}

//------------------------------------------------------------------------------
auto wrt_json_file(Json const& json, std::string const& filepath) -> bool {
  bool     res{false};
  ofstream ofs{filepath};

  try {
    ofs << json;
    res = true;
  }
  // https://json.nlohmann.me/api/basic_json/operator_ltlt/#exceptions
  catch (Json::type_error const& e) {
    cerr << "\n  Json::type_error \n'" << e.what()
         << "'\n  caught in wrt_json_file()\n\n";

    // DESIGN: report stream position here as well?

    if (e.id == 316)
      cerr << "'string stored inside the JSON value is not UTF-8 encoded'\n"
           << "https://json.nlohmann.me/home/exceptions/"
              "#jsonexceptiontype_error316\n";

    // ofs.clear(); // DESIGN: TODO: chk was the file corrupted ?

  } catch (exception const& e) {
    cerr << "\n  exception '" << e.what() << "' caught in wrt_json_file()\n\n";

    // ofs.clear(); // DESIGN: TODO: chk was the file corrupted ?
  }

  ofs.close();
  return res;
}

//------------------------------------------------------------------------------
template <typename T>
auto try_emplace_back(Json& json, T const& typ) -> bool {
  bool res{false};

  try {
    json.emplace_back(typ);
    res = true;
  }
  // https://json.nlohmann.me/api/basic_json/emplace_back/#exceptions
  catch (Json::type_error const& e) {
    cerr << "\n  Json::type_error \n'" << e.what()
         << "'\n  caught in try_emplace_back()\n\n";

    if (e.id == 311)
      cerr << "'type is other than JSON array or null'\n"
           << "https://json.nlohmann.me/home/exceptions/"
              "#jsonexceptiontype_error311\n";

  } catch (exception const& e) {
    cerr << "\n  exception '" << e.what()
         << "' caught in try_emplace_back()\n\n";
  }

  return res;
}

//------------------------------------------------------------------------------
auto try_idx_tag(Json::iterator idx_it, std::string const& tag) -> Json& {  //

  try {
    return idx_it->at(tag);
  }
  // https://json.nlohmann.me/api/basic_json/at/#exceptions
  catch (Json::type_error const& e) {
    cerr << "\n  Json::type_error \n'" << e.what()
         << "'\n  caught in try_idx_tag()\n\n";

    if (e.id == 304)
      cerr << "'JSON value is not an array'\n"
           << "https://json.nlohmann.me/home/exceptions/"
              "#jsonexceptiontype_error304\n";

  } catch (Json::out_of_range const& e) {
    cerr << "\n  Json::out_of_range \n'" << e.what()
         << "'\n  caught in try_idx_tag()\n\n";

    if (e.id == 403)
      cerr << "'key is not stored in the object'\n"
           << "https://json.nlohmann.me/home/exceptions/"
              "#jsonexceptionout_of_range403\n";

  } catch (exception const& e) {
    cerr << "\n  exception '" << e.what() << "' caught in try_idx_tag()\n\n";
  }

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wreturn-type"
}
#pragma GCC diagnostic pop

}  // namespace rw::json

//    --=^=--

// "So is it true Sempai? Are robowaifus really better than any human woman?"

// rw_foundations_lib.v220312
// Copyright (2022)
// License MIT(Expat) https://opensource.org/licenses/MIT
